from __future__ import nested_scopes

import os, sys, string, xml.sax, ConfigParser, random

import xmlparsers

# used to hold Configuration Instance
config = ''
installed = ''

__version__ = "$Revision: 1.12 $"

def setConf(input):
    """Handle configuration/setup"""

    config = getConfig()
    if len(input) > 1:
        if input[1] == 'set' and len(input) > 3:
            config.setParameter(input[2], input[3])
        elif input[1] == 'setall':
            config.configAllParameters()
        elif input[1] == 'list':
            params = config.returnAllParameters()
            sys.stdout.write('Configuration parameters\n')
            for x in params:
                sys.stdout.write('%s = %s\n' % (x[0], x[1]))
            return
        elif input[1] == 'write':
            config.writeConfig()
        else:
            sys.stdout.write("Invalid options, check help\n")
    else:
        sys.stdout.write("Invalid options, check help\n")

def listpkg(input):
    """List installed packages. If input is empty list all packages, if
    input[0] has a string check to see if there is a package with that
    name installed."""

    packages = InstalledPackages()
    if len(input) > 1:
        installed = packages.installed(input[1])
        if installed:
            sys.stdout.write("Package %s is installed\n" % input[1])
        else:
            sys.stdout.write("Package %s is not installed\n" % input[1])
    else:
        packageList = packages.returnAll()
        sys.stdout.write("Installed Packages: \n")
        for pkg in packageList.keys():
            sys.stdout.write("%s \n" % pkg)

def installModule(commands, help):
    """Install the config command into the main eval loop"""

    commands['config'] = setConf
    commands['list'] = listpkg
    help['config'] = 'config (set|list|write|setall) to set a parameter, list parameter values, or \nwrite configuration file'
    help['list'] = 'list [package] to see if package is installed. list to see all installed packages'



class Configuration:
    """Holds configuration information"""

    def __init__(self):
        """Initialize configuration, prompting the user if necessary"""

        # right now this doesn't do much, but eventually this will
        # read configuration files and do a bunch of other things
        self.__servers = []
        self.__parameters = {}
        self.__parser = ConfigParser.ConfigParser({'readline': 'false'})
        random.seed()
        try:
            self.__parser.read(os.path.expanduser('~/.ciphonrc'))
        except ParsingError:
            raise ConfigError()
        

        if not self.__parser.has_section('Ciphon'):
            self.__parser.add_section('Ciphon')
            self.configAllParameters()

        try:
            serverParser = xmlparsers.ServerReader()
            xml.sax.parse(os.path.expanduser('~/.ciphon/servers.xml'),
                          serverParser)
            self.__servers = serverParser.getServerList()
            
            # this is used to hold a working copy of the server list
            self.__downloadServers = self.__servers
        except xmlparsers.InvalidXML, x:
            sys.stderr.write("Error parsing installed.xml: %s" % x)


    def writeConfig(self):
        """Write config file to ~/.ciphonrc"""

        try:
            configFile = open(os.path.expanduser('~/.ciphonrc'), 'w')
        except IOError:
            sys.stdout.write("Error writing ~/.ciphonrc\n")
            return
        self.__parser.write(configFile)
        configFile.close()

    def setParameter(self, parameter, value):
        """Set parameter to value"""

        if parameter == 'installtype':
            if value == 'rpm':
                self.__parser.set('Ciphon', 'installtype', 'rpm')
            elif value == 'standard':
                self.__parser.set('Ciphon', 'installtype', 'standard')
            else:
                sys.stdout.write("Installation type must be 'rpm' or 'standard'")
        elif parameter == 'autoupdate':
            if value == 'true':
                self.__parser.set('Ciphon', 'autoupdate', 'true')
            elif value == 'false':
                self.__parser.set('Ciphon', 'autoupdate', 'false')
            else:
                sys.stdout.write("autoupdate must be 'true' or 'false'")
        elif parameter == 'readline':
            if value == 'true':
                self.__parser.set('Ciphon', 'readline', 'true')
            elif value == 'false':
                self.__parser.set('Ciphon', 'readline', 'false')
            else:
                sys.stdout.write("readline must be 'true' or 'false'")
        elif parameter == 'forceupdate':
            if value == 'true':
                self.__parser.set('Ciphon', 'forceupdate', 'true')
            elif value == 'false':
                self.__parser.set('Ciphon', 'forceupdate', 'false')
            else:
                sys.stdout.write("forceupdate must be 'true' or 'false'")
        else:
            sys.stdout.write("Invalid configuration parameter\n")

    def getParameter(self, parameter):
        """Get parameter value"""

        return self.__parser.get('Ciphon', parameter)
    
    def returnAllParameters(self):
        """Return all configuration parameters as (key, value) tuples"""

        parameters = []
        for option in self.__parser.options('Ciphon'):
            parameters.append((option, self.__parser.get('Ciphon', option)))
        return parameters

    def configAllParameters(self):
        """Get user configuration of all parameters"""
        while 1:
            sys.stdout.write("Use rpms for installation [N/y]? ")
            input = sys.stdin.readline()
            input = input.strip().upper()
            if input == '' or input == 'N':
                self.setParameter('installtype', 'standard')
                break
            elif input == 'Y':
                self.setParameter('installtype', 'rpm')
                break

        while 1:
            sys.stdout.write("Update package and server list on startup [Y/n]? ")
            input = sys.stdin.readline()
            input = input.strip().upper()
            if input == '' or input == 'Y':
                self.setParameter('autoupdate', 'true')
                break
            elif input == 'N':
                self.setParameter('autoupdate', 'false')
                break

        while 1:
            sys.stdout.write("Force updates[N/y]? ")
            input = sys.stdin.readline()
            input = input.strip().upper()
            if input == '' or input == 'N':
                self.setParameter('forceupdate', 'false')
                break
            elif input == 'Y':
                self.setParameter('forceupdate', 'true')
                break
            
        while 1:
            sys.stdout.write("Use readline[N/y]? ")
            input = sys.stdin.readline()
            input = input.strip().upper()
            if input == '' or input == 'N':
                self.setParameter('readline', 'false')
                break
            elif input == 'Y':
                self.setParameter('readline', 'true')
                break
            
                
    def resetServerAddress(self):
        """Reset the server list to the entire list of servers"""

        self.__downloadServers = self.__servers

    def getServerAddress(self):
        """return next server on list

        Randomly chooses a server from self.__downloadServer, deletes that
        entry and then returns that entry"""

        max = len(self.__servers)
        if max == 0:
            raise ServerError()
        index = random.randrange(0, max, 1)
        server = self.__downloadServers[index]
        del self.__downloadServers[index]

        return server

    def getPackageType(self):
        """Return the type of package to build and install on this system"""

        if self.__parameters.has_key('package-type'):
            return self.__parameters['package-type']
        else:
            return 'standard'

    def __createConfig(self, fileName):
        """Ask user for configuration and create a configuration file"""

        sys.stdout.write('No configuration file found generating a new one.\n')
        sys.stdout.write('Enter a blank line to exit\n')
        done = 0
        while not done:
            sys.stdout.write('Enter server address: ')
            server = sys.stdin.readline()
            if len(server) == 1:
                done = 1
            else:
                self.__servers.append(server)
        try:
            rcFile = open(fileName, 'w')
        except:
            sys.stderr.write("Can't write config file, continuing without it...\n")
            return
        for server in self.__servers:
            rcFile.write('\nserver = %s' % server)
        rcFile.close()

class InstalledPackages:
    """Holds information about installed packages"""
    def __init__(self):

        try:
            installedParser = xmlparsers.InstalledReader()
            xml.sax.parse(os.path.expanduser('~/.ciphon/installed.xml'),
                          installedParser)
            self.__installed = installedParser.getInstalledList()
        except xmlparsers.InvalidXML, x:
            sys.stderr.write("Error parsing installed.xml: %s" % x)
        except (IOError, OSError):
            self.__installed = {}

    def installed(self, package):
        """Check to see if package is installed, if so return true"""
        if self.__installed.has_key(package):
            return 1
        else:
            return 0

    def current(self, package, version):
        """Check to see if given package and version is newer than installed
        version, returns 0 if passed version is newer. 

        Gets the installed version and version, splits them on ., and then
        compares the two lists element wise.  Returns 0 if version's components
        are all larger than the installed version's components.  Also returns 0
        if the package is not installed or if an error occurs"""

        if self.__installed.has_key(package):
            installedVersion = self.__installed[package]
            installedVersion = installedVersion.split('.')
            version = version.split('.')
            newer = 1
            try:
                for (i, n) in zip(installedVersion, version):
                    if int(i) < int(n):
                        newer = 0
            except ValueError:
                return 0
            if not newer and len(version) > len(installedVersion):
                # if version has more elements than installedVersion and
                # everything else is the same then version is larger
                return 0
            return newer
        else:
            return 1
                    
    def update(self, package, version):
        """Update the installed package with package and version"""

        self.__installed[package] = version

    def write(self):
        """write installed file"""

        try:
            fileHandle = open(os.path.expanduser('~/.ciphon/installed.xml'), 'w')
        except (IOError, OSError):
            sys.stderr.write("Can't write to installed.xml...giving up\n");
        fileHandle.write("<packagelist>\n")
        for package in self.__installed.keys():
            version = self.__installed[package]
            fileHandle.write('<package name="%s" version="%s" />\n' % (package,
                                                                       version))
        fileHandle.write('</packagelist>\n')
        fileHandle.close()
          
    def returnAll(self):
        """Return an array with the names of all installed packages"""
        return self.__installed

def getConfig():
    """Return an instance of Configuration

    Currently this function returns a new instance of Configuration.  Later
    on, this function should try to create a single instance of Configuration
    and try to return a reference to it.  E.g. make Configuration a singleton
    class"""
    global config
    if config:
        return config
    else:
        config = Configuration()
        return config

def getInstalled():
    """Return an instance of Configuration

    Currently this function returns a new instance of Configuration.  Later
    on, this function should try to create a single instance of Configuration
    and try to return a reference to it.  E.g. make Configuration a singleton
    class"""
    global installed
    if installed:
        return installed
    else:
        installed = InstalledPackages()
        return installed




# ************* Exceptions thrown by Installer class *****************

class ConfigError(Exception):
    """Error reading configuration file"""
    pass

class ServerError(ConfigError):
    """Indicates that the list of servers for a package has been exhausted"""
    pass
